from pathlib import Path
import pytest

from csv_dataset import (
    CsvReader,
    Dataset,
    RangeNormalizer
)

csv_path = Path(Path(__file__).resolve()).parent.parent / \
    'example' / 'stock.csv'


def test_reader_exception():
    with pytest.raises(
        ValueError,
        match='positive'
    ):
        CsvReader(
            csv_path.absolute(),
            float,
            [
                2, 3, 4, 5, 6
            ],
            max_lines=0
        )


def test_window_and_batch():
    dataset = Dataset(
        CsvReader(
            csv_path.absolute(),
            float,
            [
                2, 3, 4, 5, 6
            ],
            header=True
        )
    ).window(5, 1).batch(5)

    assert dataset.lines_need(reads=3) == 19

    data = dataset.get()

    with pytest.raises(
        RuntimeError,
        match='forbidden'
    ):
        dataset.window(5, 2)

    assert len(data) == 5
    assert len(data[0]) == 5
    assert data[0][0][0] == 7145.99
    assert data[0][1][0] == 7142.89
    assert data[1][0][0] == 7142.89

    data = dataset.get()
    assert data[0][0][0] == 7134.99


def test_iterator():
    dataset = Dataset(
        CsvReader(
            csv_path.absolute(),
            float,
            [
                2, 3, 4, 5, 6
            ],
            header=True
        )
    ).window(5, 1).batch(5)

    assert len(list(dataset)) == 19

    dataset.reset()
    assert len(list(dataset)) == 19


def test_direct_get():
    dataset = Dataset(
        CsvReader(
            csv_path.absolute(),
            float,
            [
                2, 3, 4, 5, 6
            ],
            header=True
        )
    )

    got = dataset.get()

    assert got.shape == (5,)

    assert list(got) == [
        7145.99, 7150.0, 7141.01, 7142.33, 21.094283
    ]


def test_window_only():
    dataset = Dataset(
        CsvReader(
            csv_path.absolute(),
            float,
            [
                2, 3, 4, 5, 6
            ],
            header=True
        )
    ).window(5, 1)

    assert dataset.get().shape == (5, 5)


def test_batch_only():
    dataset = Dataset(
        CsvReader(
            csv_path.absolute(),
            float,
            [
                2, 3, 4, 5, 6
            ],
            header=True
        )
    ).batch(5)

    got = dataset.get()

    assert got[0][0] == 7145.99
    assert got[1][0] == 7142.89


def test_reader_with_normalizers():
    price_normalizer = RangeNormalizer(0, 10000)
    volume_normalizer = RangeNormalizer(0, 200)

    dataset = Dataset(
        CsvReader(
            csv_path.absolute(),
            float,
            [
                2, 3, 4, 5, 6
            ],
            normalizers=[
                price_normalizer,
                price_normalizer,
                price_normalizer,
                price_normalizer,
                volume_normalizer
            ],
            header=True
        )
    ).batch(5)

    got = dataset.get()

    assert format(got[0][0], '.6f') == '0.714599'
    assert format(got[1][0], '.6f') == '0.714289'

    assert format(
        price_normalizer.restore(got[0][0]),
        '.2f'
    ) == '7145.99'


def test_reader_error():
    with pytest.raises(
        ValueError,
        match='expect 5 but got 2'
    ):
        CsvReader(
            'fake',
            float,
            indexes=[1, 2, 3, 4, 5],
            normalizers=[1, 2]
        )


def test_max_lines():
    reader = CsvReader(
        csv_path.absolute(),
        float,
        [
            2, 3, 4, 5, 6
        ],
        header=True,
        max_lines=5
    )

    data = Dataset(reader)

    assert len(list(data)) == 5

    reader.max_lines = 1
    data.reset()

    assert len(list(data)) == 1


def test_read():
    reader = CsvReader(
        csv_path.absolute(),
        float,
        [
            2, 3, 4, 5, 6
        ],
        header=True
    )

    assert reader.max_lines is None

    data = Dataset(reader).window(2, shift=1)

    assert data.get()[0][0] == 7145.99

    assert reader.lines == 2

    array = data.read(4)
    assert len(array) == 4
    assert array[0][0][0] == 7142.89

    assert reader.lines == 6

    # reset

    data.reset()

    assert data.get()[0][0] == 7145.99

    array = data.read(4, reset_buffer=True)
    assert len(array) == 4
    assert array[0][0][0] == 7125.76

    assert reader.lines == 7


def test_max_reads():
    reader = CsvReader(
        csv_path.absolute(),
        float,
        [
            2, 3, 4, 5, 6
        ],
        header=True
    )

    data = Dataset(reader).window(2, shift=1)

    assert data.max_reads() is None

    with pytest.raises(ValueError, match='positive'):
        data.max_reads(-1)

    assert data.max_reads(1) == 0
    assert data.max_reads(2) == 1
    assert data.max_reads(3) == 2



from typing import (
    Optional
)

import numpy as np


def rolling_window(
    array: np.ndarray,
    size: int,
    shift: Optional[int] = None,
    stride: int = 1
) -> np.ndarray:
    """Gets an `size`-period rolling window for `array` as an 1d array

    |-------- size:3 --------|
    |- stride:1 -|           |
    |            |           |
    1            2           3 --------|---
                                    shift:2
    3            4           5 --------|---

    5            6           7
    """

    shift = shift or size

    window_step = shift * stride

    step_length = len(array) - size
    steps = step_length // window_step
    rest = step_length - window_step * steps

    steps += 1

    if rest:
        # drop the last window
        # if its size is smaller than `size`.
        array = array[:-rest]

    item_stride = array.strides[0]

    ret = np.lib.stride_tricks.as_strided(
        array,
        # rolling_window "destroy" the first dimension of input `array`,
        shape=(steps, size) + array.shape[1:],
        strides=(item_stride * shift, item_stride * stride) + array.strides[1:]
    )

    return ret


def max_lines_error(max_lines):
    return ValueError(
        f'max_lines must be positive, but got `{max_lines}`'
    )



from typing import (
    Protocol
)


class NormalizerProtocol(Protocol):
    def normalize(self, value: float) -> float:
        ...  # pragma: no cover

    def restore(self, value: float) -> float:
        ...  # pragma: no cover


class RangeNormalizer:
    def __init__(
        self,
        min_value: float,
        max_value: float
    ):
        self._min = min_value
        self._scale = max_value - min_value

    def normalize(self, value: float) -> float:
        return (value - self._min) / self._scale

    def restore(self, value: float) -> float:
        return value * self._scale + self._min



from typing import (
    List,
    Generic,
    TypeVar,
    Type,
    Optional,
    TextIO
)
from abc import (
    ABC,
    abstractmethod
)

from common_decorators import lazy

from .normalizer import NormalizerProtocol
from .common import max_lines_error


SPLITTER = ','
T = TypeVar('T', float, int)


class AbstractReader(ABC, Generic[T]):
    dtype: Type[T]

    _max_lines: Optional[int]
    _max_lines = None

    @property
    def max_lines(self) -> Optional[int]:
        return self._max_lines

    @max_lines.setter
    def max_lines(self, max_lines: Optional[int]) -> None:
        self._set_max_lines(max_lines)

    def _set_max_lines(self, max_lines: Optional[int]) -> None:
        """Set the max_lines of the reader

        Args:
            max_lines (int | None): if None, it indicates where is no limit
        """

        if max_lines is None:
            self._max_lines = None
            return

        if max_lines <= 0:
            raise max_lines_error(max_lines)

        self._max_lines = max_lines

    @abstractmethod
    def readline(_) -> List[T]:
        """Reads a single line and returns the content
        """
        ...  # pragma: no cover

    @abstractmethod
    def reset(_) -> None:
        """Reset the reader pos
        """
        ...  # pragma: no cover


class CsvReader(AbstractReader[T]):
    def __init__(
        self,
        filepath: str,
        dtype: Type[T],
        indexes: List[int],
        header: bool = False,
        splitter: str = SPLITTER,
        normalizers: List[NormalizerProtocol] = [],
        max_lines: Optional[int] = None
    ):
        self.dtype = dtype
        self._splitter = splitter
        self._indexes = indexes
        self._filepath = filepath
        self._normalizers = normalizers
        self._header = header

        self.max_lines = max_lines

        if normalizers and len(normalizers) != len(indexes):
            raise ValueError(
                f'normalizers has different length with indexes, expect {len(indexes)} but got {len(normalizers)}'
            )

        self.reset()

    @property
    def lines(self) -> int:
        """How many lines the reader has read
        """

        return self._lines

    # @property
    # def pos(self) -> int:
    #     """The stream position of the reader
    #     """

    #     return self._pos

    @lazy
    def _fd(self) -> TextIO:
        # Ref
        # https://docs.python.org/3/library/io.html
        return open(self._filepath, 'r')

    def reset(self) -> None:
        self._lines = 0
        self._fd.seek(0)

        if self._header:
            self._readline()

    def _readline(self) -> str:
        return self._fd.readline().strip()

    def _normalize(
        self,
        data: list
    ) -> list:
        if not self._normalizers:
            return data

        return [
            self._normalizers[i].normalize(datum)
            for i, datum in enumerate(data)
        ]

    def readline(self) -> Optional[List[T]]:
        if self._lines == self._max_lines:
            return

        line = self._readline()
        # self._pos = self._fd.tell()

        if not line:
            return

        splitted = line.split(self._splitter)

        try:
            line = [
                self.dtype(cell)
                for i, cell in enumerate(splitted)
                if i in self._indexes
            ]
        except ValueError:
            return self.readline()

        self._lines += 1

        return self._normalize(line)



import numpy as np

from csv_dataset.common import rolling_window


def test_rolling_window_1d():
    expected = np.array([
        np.arange(0, 3),
        np.arange(3, 6),
        np.arange(6, 9),
    ])

    assert np.array_equal(
        rolling_window(
            np.arange(10),
            3
        ), expected
    )

    assert np.array_equal(
        rolling_window(
            np.arange(11),
            3
        ), expected
    )


def test_rolling_window_1d_with_shift():
    expected = np.array([
        [0, 1, 2],
        [1, 2, 3],
        [2, 3, 4]
    ])

    assert np.array_equal(
        rolling_window(
            np.arange(5),
            3,
            shift=1
        ), expected
    )


def test_rolling_window_1d_with_shift_and_stride():
    expected = np.array([
        [0, 2, 4],
        [1, 3, 5],
        [2, 4, 6]
    ])

    assert np.array_equal(
        rolling_window(
            np.arange(7),
            3,
            shift=1,
            stride=2
        ), expected
    )


def test_rolling_window_2d():
    def create(length):
        return rolling_window(
            rolling_window(
                np.arange(length),
                2
            ),
            2
        )

    expected = np.array([
        [
            [0, 1],
            [2, 3]
        ],
        [
            [4, 5],
            [6, 7]
        ]
    ])

    np.testing.assert_array_equal(create(8), expected)
    np.testing.assert_array_equal(create(9), expected)
    np.testing.assert_array_equal(create(10), expected)


def test_rolling_window_2d_with_shift_and_stride():
    def create(length, **kwargs):
        return rolling_window(
            # [
            #     [0, 1],
            #     [2, 3],
            #     [4, 5],
            #     [6, 7]
            # ]
            rolling_window(
                np.arange(length),
                2
            ),
            2,
            **kwargs
        )

    np.testing.assert_array_equal(create(8, shift=1), np.array([
        [
            [0, 1], [2, 3]
        ],
        [
            [2, 3], [4, 5]
        ],
        [
            [4, 5], [6, 7]
        ]
    ]))

    np.testing.assert_array_equal(create(8, shift=1, stride=2), np.array([
        [
            [0, 1], [4, 5]
        ],
        [
            [2, 3], [6, 7]
        ]
    ]))



from typing import (
    Optional,
    TypeVar,
    Generic
)

import numpy as np

from common_decorators import lazy

from .reader import AbstractReader
from .common import (
    rolling_window,
    max_lines_error
)

T = TypeVar('T')


class Dataset(Generic[T]):
    _reader: AbstractReader[T]

    def __init__(
        self,
        reader: AbstractReader[T]
    ):
        self._reader = reader
        self._buffer_read = False

        self._batch = 1

        self._window_size = 1
        self._window_shift = 1
        self._window_stride = 1
        self._window_drop_remainder = False

        self.reset()

    def reset(self) -> None:
        self._reader.reset()
        self._buffer = None

        return self

    # |-------- size:3 --------|
    # |- stride:1 -|           |
    # |            |           |
    # 1            2           3 --------|---
    #                                 shift:2
    # 3            4           5 --------|---
    #
    # 5            6           7
    #
    # least: (size - 1) * stride + 1
    # step : shift * stride
    @lazy
    def _single_least(self) -> int:
        """The least size (lines) of the underlying datum referred to the first single window
        """

        return (self._window_size - 1) * self._window_stride + 1

    @lazy
    def _single_stride(self) -> int:
        """stride without considering the batch size
        """
        return self._window_shift * self._window_stride

    @lazy
    def _least(self) -> int:
        """The least size (lines) of the underlying datum referered to the first batch read
        """

        return self._single_least + (self._batch - 1) * self._single_stride

    @lazy
    def _stride(self) -> int:
        """stride that considering window stride and batch size
        """

        return self._batch * self._single_stride

    def lines_need(self, reads: int) -> int:
        """Calculate how many lines of datum needed for reading `reads` times
        """

        return self._least + (reads - 1) * self._stride

    def max_reads(self, max_lines: Optional[int] = None) -> Optional[int]:
        """How many reads does the current dataset afford

        Args:
            max_lines (:obj:`int`, optional)

        Returns:
            - None which indicates the dataset could not know about it. If max_lines is not provided, and the reader has no max_lines
            - int otherwise
        """

        if max_lines is None:
            max_lines = self._reader.max_lines

        if max_lines is None:
            return None

        if max_lines <= 0:
            raise max_lines_error(max_lines)

        rest = max_lines - self._least

        return 0 if rest < 0 else 1 + rest // self._stride

    def _check_start(self, method_name: str):
        if self._buffer_read:
            raise RuntimeError(f'calling {method_name}() during reading is forbidden')

    def window(
        self,
        size: int,
        shift: Optional[int] = None,
        stride: int = 1
    ) -> 'Dataset':
        """Combines (nests of) input elements into a dataset of (nests of) windows.
        """

        self._check_start('window')

        self._window_size = size
        self._window_shift = shift or size
        self._window_stride = stride

        return self

    def batch(
        self,
        batch: int
    ) -> 'Dataset':
        """Combines consecutive elements of this dataset into batches.
        """

        self._check_start('batch')

        self._batch = batch

        return self

    def _readlines(
        self,
        lines: int,
        dest_buffer: list,
        slice_size: bool = False
    ) -> list:
        read = 0

        while read < lines:
            line = self._reader.readline()

            if line is None:
                # Reaches EOF
                return []

            dest_buffer.append(line)
            read += 1

        if slice_size:
            dest_buffer = dest_buffer[lines:]

        return dest_buffer

    def _read_buffer(self):
        if self._buffer is None:
            # Initialize buffer
            self._buffer_read = True
            self._buffer = self._readlines(self._least, [])
        else:
            # Extend buffer
            self._buffer = self._readlines(self._stride, self._buffer, True)

    def get(self) -> Optional[np.ndarray]:
        """Gets the data of the next batch
        """

        self._read_buffer()

        if not self._buffer:
            # There is no data,
            # which indicates that the data has been exhausted
            return

        array = np.array(self._buffer)

        windowed = array if self._window_size == 1 else rolling_window(
            array,
            self._window_size,
            shift=self._window_shift,
            stride=self._window_stride
        )

        batched = windowed if self._batch == 1 else rolling_window(
            windowed,
            self._batch
        )

        return batched[0]

    def reset_buffer(self) -> None:
        self._buffer = None

    def read(
        self,
        amount: int,
        reset_buffer: bool = False
    ) -> list:
        """Reads multiple batches of data

        Args:
            amount (int): number of batches to be read
            reset_buffer (:obj:`bool`, optional): if `True`, the dataset will reset the data of the previous window in the buffer
        """

        if reset_buffer:
            self.reset_buffer()

        array = []

        while amount > 0:
            amount -= 1
            array.append(self.get())

        return array

    def __iter__(self) -> 'Dataset':
        return self

    def __next__(self) -> np.ndarray:
        got = self.get()

        if got is None:
            raise StopIteration

        return got
